@@ -7,17 +7,16 @@
 #define GPDATADIR "/usr/local/share/pari"
 #define SHELL_Q '\''
 
-#define PARIVERSION "GP/PARI CALCULATOR Version 2.15.5 (released)"
-#define PARIINFO "powerpc64 running openbsd (portable C/GMP-%s kernel) 64-bit version"
-#define PARI_VERSION_CODE 134917
+#define PARIVERSION "GP/PARI CALCULATOR Version 2.17.2 (released)"
+#define PARIINFO "amd64 running openbsd (x86-64/GMP-%s kernel) 64-bit version"
+#define PARI_VERSION_CODE 135426
 #define PARI_VERSION(a,b,c) (((a) << 16) + ((b) << 8) + (c))
 #define PARI_VERSION_SHIFT 8
 #define PARI_VCSVERSION ""
 #define PARI_MT_ENGINE "single"
 
 #define PARI_DOUBLE_FORMAT -
-#define GCC_VERSION "OpenBSD clang version 16.0.6"
-#define __HAS_NO_ASM__
+#define GCC_VERSION "OpenBSD clang version 19.1.7"
 #define ASMINLINE
 
 /*  Location of GNU gzip program (enables reading of .Z and .gz files). */
@@ -29,6 +28,7 @@
 
 #define READLINE "4.3"
 #define LONG_IS_64BIT
+#define HAS_SSE2
 #define HAS_EXP2
 #define HAS_LOG2
 #define HAS_RINT
