@@ -4,13 +4,15 @@ This file is part of the PARI/GP package.
 
 PARI/GP is free software; you can redistribute it and/or modify it under the
 terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
+Foundation; either version 2 of the License, or (at your option) any later
+version. It is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
+
 /*********************************************************************/
 /*                       MALLOC/FREE WRAPPERS                        */
 /*********************************************************************/
@@ -74,13 +76,8 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
        gel(_v,1) = (x);\
        gel(_v,2) = (y);\
        gel(_v,3) = (z); return _v; } while(0)
-#define retmkqfi(x,y,z)\
-  do { GEN _v = cgetg(4, t_QFI);\
-       gel(_v,1) = (x);\
-       gel(_v,2) = (y);\
-       gel(_v,3) = (z); return _v; } while(0)
-#define retmkqfr(x,y,z,d)\
-  do { GEN _v = cgetg(5, t_QFR);\
+#define retmkqfb(x,y,z,d)\
+  do { GEN _v = cgetg(5, t_QFB);\
        gel(_v,1) = (x);\
        gel(_v,2) = (y);\
        gel(_v,3) = (z);\
@@ -161,7 +158,10 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
        gel(_v,3) = (z);\
        gel(_v,4) = (t);\
        gel(_v,5) = (u); return _v; } while(0)
-
+#define retmkmat22(a,b,c,d)\
+  do { GEN _v = cgetg(3, t_MAT);\
+       gel(_v,1) = mkcol2(a,c);\
+       gel(_v,2) = mkcol2(b,d); return _v; } while(0)
 INLINE GEN
 mkintmod(GEN x, GEN y) { retmkintmod(x,y); }
 INLINE GEN
@@ -175,7 +175,7 @@ mkpolmod(GEN x, GEN y) { retmkpolmod(x,y); }
 INLINE GEN
 mkfrac(GEN x, GEN y) { retmkfrac(x,y); }
 INLINE GEN
-mkfracss(long x, long y) { retmkfrac(stoi(x),stoi(y)); }
+mkfracss(long x, long y) { retmkfrac(stoi(x),utoipos(y)); }
 /* q = n/d a t_FRAC or t_INT; recover (n,d) */
 INLINE void
 Qtoss(GEN q, long *n, long *d)
@@ -195,11 +195,32 @@ sstoQ(long n, long d)
   }
   if (d < 0) { d = -d; n = -n; }
   if (d == 1) return stoi(n);
-  q = udivuu_rem(labs(n),d,&r);
+  r = labs(n);
+  if (r == 1) retmkfrac(n > 0? gen_1: gen_m1, utoipos(d));
+  q = udivuu_rem(r, d, &r);
   if (!r) return n > 0? utoipos(q): utoineg(q);
   g = ugcd(d,r); /* gcd(n,d) */
   if (g != 1) { n /= g; d /= g; }
-  retmkfrac(stoi(n), utoi(d));
+  retmkfrac(stoi(n), utoipos(d));
+}
+
+INLINE GEN
+uutoQ(ulong n, ulong d)
+{
+  ulong r;
+  long g, q;
+  if (!n)
+  {
+    if (!d) pari_err_INV("uutoQ",gen_0);
+    return gen_0;
+  }
+  if (d == 1) return utoipos(n);
+  if (n == 1) retmkfrac(gen_1, utoipos(d));
+  q = udivuu_rem(n,d,&r);
+  if (!r) return utoipos(q);
+  g = ugcd(d,r); /* gcd(n,d) */
+  if (g != 1) { n /= g; d /= g; }
+  retmkfrac(utoipos(n), utoipos(d));
 }
 
 INLINE GEN
@@ -236,7 +257,7 @@ mkvecsmall5(long x,long y,long z,long t,long u) { GEN v = cgetg(6, t_VECSMALL);
   v[1]=x; v[2]=y; v[3]=z; v[4]=t; v[5]=u; return v; }
 
 INLINE GEN
-mkqfi(GEN x, GEN y, GEN z) { retmkqfi(x,y,z); }
+mkqfb(GEN x, GEN y, GEN z, GEN d) { retmkqfb(x,y,z,d); }
 /* vec */
 INLINE GEN
 mkvec(GEN x) { retmkvec(x); }
@@ -334,6 +355,15 @@ pol_xnall(long n, long v)
   if (n < 0) retmkrfrac(gen_1, pol_xn(-n,v));
   return pol_xn(n, v);
 }
+/* x^n, assume n >= 0 */
+INLINE GEN
+polxn_Flx(long n, long sv) {
+  long i, a = n+2;
+  GEN p = cgetg(a+1, t_VECSMALL);
+  p[1] = sv;
+  for (i = 2; i < a; i++) p[i] = 0;
+  p[a] = 1; return p;
+}
 INLINE GEN
 pol_1(long v) {
   GEN p = cgetg(3, t_POL);
@@ -396,7 +426,7 @@ INLINE GEN
 zeroser(long v, long e)
 {
   GEN x = cgetg(2, t_SER);
-  x[1] = evalvalp(e) | evalvarn(v); return x;
+  x[1] = evalvalser(e) | evalvarn(v); return x;
 }
 INLINE int
 ser_isexactzero(GEN x)
@@ -564,6 +594,32 @@ vecsmalltrunc_init(long l)
 INLINE void
 vecsmalltrunc_append(GEN x, long t) { x[ lg(x) ] = t; lg_increase(x); }
 
+/*******************************************************************/
+/*                                                                 */
+/*                    STRING HASH FUNCTIONS                        */
+/*                                                                 */
+/*******************************************************************/
+INLINE ulong
+hash_str(const char *str)
+{
+  ulong hash = 5381UL, c;
+  while ( (c = (ulong)*str++) )
+    hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
+  return hash;
+}
+INLINE ulong
+hash_str_len(const char *str, long len)
+{
+  ulong hash = 5381UL;
+  long i;
+  for (i = 0; i < len; i++)
+  {
+    ulong c = (ulong)*str++;
+    hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
+  }
+  return hash;
+}
+
 /*******************************************************************/
 /*                                                                 */
 /*                        VEC / COL / VECSMALL                     */
@@ -573,19 +629,18 @@ vecsmalltrunc_append(GEN x, long t) { x[ lg(x) ] = t; lg_increase(x); }
 INLINE GEN
 vec_shorten(GEN v, long n)
 {
+  GEN V = cgetg(n+1, t_VEC);
   long i;
-  GEN V = cgetg(n+1,t_VEC);
-  for(i=1;i<=n;i++) gel(V,i) = gel(v,i);
+  for(i = 1; i <= n; i++) gel(V,i) = gel(v,i);
   return V;
 }
 /* shallow*/
 INLINE GEN
 vec_lengthen(GEN v, long n)
 {
-  long i;
-  long l=lg(v);
-  GEN V = cgetg(n+1,t_VEC);
-  for(i=1;i<l;i++) gel(V,i) = gel(v,i);
+  GEN V = cgetg(n+1, t_VEC);
+  long i, l = lg(v);
+  for(i = 1; i < l; i++) gel(V,i) = gel(v,i);
   return V;
 }
 /* shallow*/
@@ -594,7 +649,7 @@ vec_append(GEN V, GEN s)
 {
   long i, l2 = lg(V);
   GEN res = cgetg(l2+1, typ(V));
-  for (i = 1; i < l2; ++i) gel(res, i) = gel(V,i);
+  for (i = 1; i < l2; i++) gel(res, i) = gel(V,i);
   gel(res,l2) = s; return res;
 }
 /* shallow*/
@@ -604,7 +659,7 @@ vec_prepend(GEN v, GEN s)
   long i, l = lg(v);
   GEN w = cgetg(l+1, typ(v));
   gel(w,1) = s;
-  for (i = 2; i <= l; ++i) gel(w,i) = gel(v,i-1);
+  for (i = 2; i <= l; i++) gel(w,i) = gel(v,i-1);
   return w;
 }
 /* shallow*/
@@ -618,9 +673,9 @@ vec_setconst(GEN v, GEN x)
 INLINE GEN
 vecsmall_shorten(GEN v, long n)
 {
-  long i;
   GEN V = cgetg(n+1,t_VECSMALL);
-  for(i=1;i<=n;i++) V[i] = v[i];
+  long i;
+  for(i = 1; i <= n; i++) V[i] = v[i];
   return V;
 }
 INLINE GEN
@@ -628,7 +683,7 @@ vecsmall_lengthen(GEN v, long n)
 {
   long i, l = lg(v);
   GEN V = cgetg(n+1,t_VECSMALL);
-  for(i=1;i<l;i++) V[i] = v[i];
+  for(i = 1; i < l; i++) V[i] = v[i];
   return V;
 }
 
@@ -798,6 +853,8 @@ RgX_isscalar(GEN x)
 }
 INLINE long
 RgX_equal_var(GEN x, GEN y) { return varn(x) == varn(y) && RgX_equal(x,y); }
+INLINE GEN
+RgX_to_RgV(GEN x, long N) { x = RgX_to_RgC(x, N); settyp(x, t_VEC); return x; }
 
 INLINE int
 RgX_is_rational(GEN x)
@@ -848,6 +905,16 @@ RgV_is_QV(GEN x)
     if (!is_rational_t(typ(gel(x,i)))) return 0;
   return 1;
 }
+INLINE long
+RgV_isin_i(GEN v, GEN x, long n)
+{
+  long i;
+  for (i = 1; i <= n; i++)
+    if (gequal(gel(v,i), x)) return i;
+  return 0;
+}
+INLINE long
+RgV_isin(GEN v, GEN x) { return RgV_isin_i(v, x, lg(v)-1); }
 
 /********************************************************************/
 /**                                                                **/
@@ -879,9 +946,7 @@ pari_stack_alloc(pari_stack *s, long nb)
   {
     while (s->n+nb > alloc) alloc <<= 1;
   }
-  BLOCK_SIGINT_START
-  *sdat = pari_realloc(*sdat,alloc*s->size);
-  BLOCK_SIGINT_END
+  pari_realloc_ip(sdat,alloc*s->size);
   s->alloc = alloc;
 }
 
@@ -1022,6 +1087,9 @@ INLINE GEN
 vecpermute(GEN A, GEN x)
 { pari_APPLY_type(typ(A), gel(A, x[i])) }
 
+INLINE GEN
+veclast(GEN A) { return gel(A, lg(A)-1); }
+
 INLINE GEN
 rowpermute(GEN x, GEN p)
 { pari_APPLY_same(typ(gel(x,i)) == t_VECSMALL ? vecsmallpermute(gel(x, i), p)
@@ -1032,15 +1100,25 @@ rowpermute(GEN x, GEN p)
 /*                          PERMUTATIONS                           */
 /*                                                                 */
 /*******************************************************************/
+INLINE GEN
+identity_zv(long n)
+{
+  GEN v = cgetg(n+1, t_VECSMALL);
+  long i;
+  for (i = 1; i <= n; i++) v[i] = i;
+  return v;
+}
+INLINE GEN
+identity_ZV(long n)
+{
+  GEN v = cgetg(n+1, t_VEC);
+  long i;
+  for (i = 1; i <= n; i++) gel(v,i) = utoipos(i);
+  return v;
+}
 /* identity permutation */
 INLINE GEN
-identity_perm(long n)
-{
-  GEN perm = cgetg(n+1, t_VECSMALL);
-  long i;
-  for (i = 1; i <= n; i++) perm[i] = i;
-  return perm;
-}
+identity_perm(long n) { return identity_zv(n); }
 
 /* assume d <= n */
 INLINE GEN
@@ -1058,6 +1136,10 @@ INLINE GEN
 perm_mul(GEN s, GEN x)
 { pari_APPLY_long(s[x[i]]) }
 
+INLINE GEN
+perm_sqr(GEN x)
+{ pari_APPLY_long(x[x[i]]) }
+
 /* Compute the inverse (reciprocal) of a permutation. */
 INLINE GEN
 perm_inv(GEN x)
@@ -1110,6 +1192,18 @@ pari_realloc(void *pointer, size_t size)
   if (!tmp) pari_err(e_MEM);
   return tmp;
 }
+INLINE void
+pari_realloc_ip(void **pointer, size_t size)
+{
+  char *tmp;
+  BLOCK_SIGINT_START;
+  if (!*pointer) tmp = (char *) malloc(size);
+  else tmp = (char *) realloc(*pointer,size);
+  if (!tmp) pari_err(e_MEM);
+  *pointer = tmp;
+  BLOCK_SIGINT_END;
+}
+
 INLINE void*
 pari_calloc(size_t size)
 {
@@ -1117,7 +1211,7 @@ pari_calloc(size_t size)
   memset(t, 0, size); return t;
 }
 INLINE GEN
-cgetalloc(long t, size_t l)
+cgetalloc(size_t l, long t)
 { /* evallg may raise e_OVERFLOW, which would leak x */
   ulong x0 = evaltyp(t) | evallg(l);
   GEN x = (GEN)pari_malloc(l * sizeof(long));
@@ -1129,7 +1223,7 @@ cgetalloc(long t, size_t l)
 /*                       GARBAGE COLLECTION                        */
 /*                                                                 */
 /*******************************************************************/
-/* copy integer x as if we had avma = av */
+/* copy integer x as if we had set_avma(av) */
 INLINE GEN
 icopy_avma(GEN x, pari_sp av)
 {
@@ -1139,7 +1233,7 @@ icopy_avma(GEN x, pari_sp av)
   y[0] = evaltyp(t_INT)|evallg(lx);
   return y;
 }
-/* copy leaf x as if we had avma = av */
+/* copy leaf x as if we had set_avma(av) */
 INLINE GEN
 leafcopy_avma(GEN x, pari_sp av)
 {
@@ -1155,24 +1249,24 @@ gerepileuptoleaf(pari_sp av, GEN x)
   long lx;
   GEN q;
 
-  if (!isonstack(x) || (GEN)av<=x) { avma = av; return x; }
+  if (!isonstack(x) || (GEN)av<=x) return gc_const(av,x);
   lx = lg(x);
   q = ((GEN)av) - lx;
-  avma = (pari_sp)q;
+  set_avma((pari_sp)q);
   while (--lx >= 0) q[lx] = x[lx];
   return q;
 }
 INLINE GEN
 gerepileuptoint(pari_sp av, GEN x)
 {
-  if (!isonstack(x) || (GEN)av<=x) { avma = av; return x; }
-  avma = (pari_sp)icopy_avma(x, av);
+  if (!isonstack(x) || (GEN)av<=x) return gc_const(av,x);
+  set_avma((pari_sp)icopy_avma(x, av));
   return (GEN)avma;
 }
 INLINE GEN
 gerepileupto(pari_sp av, GEN x)
 {
-  if (!isonstack(x) || (GEN)av<=x) { avma = av; return x; }
+  if (!isonstack(x) || (GEN)av<=x) return gc_const(av,x);
   switch(typ(x))
   { /* non-default = !is_recursive_t(tq) */
     case t_INT: return gerepileuptoint(av, x);
@@ -1192,21 +1286,26 @@ gerepilecopy(pari_sp av, GEN x)
   if (is_recursive_t(typ(x)))
   {
     GENbin *p = copy_bin(x);
-    avma = av; return bin_copy(p);
+    set_avma(av); return bin_copy(p);
   }
   else
   {
-    avma = av;
+    set_avma(av);
     if (x < (GEN)av) {
       if (x < (GEN)pari_mainstack->bot) new_chunk(lg(x));
       x = leafcopy_avma(x, av);
-      avma = (pari_sp)x;
+      set_avma((pari_sp)x);
     } else
       x = leafcopy(x);
     return x;
   }
 }
 
+INLINE void
+guncloneNULL(GEN x) { if (x) gunclone(x); }
+INLINE void
+guncloneNULL_deep(GEN x) { if (x) gunclone_deep(x); }
+
 /* Takes an array of pointers to GENs, of length n. Copies all
  * objects to contiguous locations and cleans up the stack between
  * av and avma. */
@@ -1215,7 +1314,7 @@ gerepilemany(pari_sp av, GEN* gptr[], int n)
 {
   int i;
   for (i=0; i<n; i++) *gptr[i] = (GEN)copy_bin(*gptr[i]);
-  avma = av;
+  set_avma(av);
   for (i=0; i<n; i++) *gptr[i] = bin_copy((GENbin*)*gptr[i]);
 }
 
@@ -1229,7 +1328,7 @@ gerepileall(pari_sp av, int n, ...)
     GEN *gptr[10];
     for (i=0; i<n; i++)
     { gptr[i] = va_arg(a,GEN*); *gptr[i] = (GEN)copy_bin(*gptr[i]); }
-    avma = av;
+    set_avma(av);
     for (--i; i>=0; i--) *gptr[i] = bin_copy((GENbin*)*gptr[i]);
 
   }
@@ -1238,19 +1337,32 @@ gerepileall(pari_sp av, int n, ...)
     GEN **gptr = (GEN**)  pari_malloc(n*sizeof(GEN*));
     for (i=0; i<n; i++)
     { gptr[i] = va_arg(a,GEN*); *gptr[i] = (GEN)copy_bin(*gptr[i]); }
-    avma = av;
+    set_avma(av);
     for (--i; i>=0; i--) *gptr[i] = bin_copy((GENbin*)*gptr[i]);
     pari_free(gptr);
   }
   va_end(a);
 }
 
+/* assume 1 <= n < 10 */
+INLINE GEN
+gc_all(pari_sp av, int n, ...)
+{
+  int i;
+  GEN *v[10];
+  va_list a; va_start(a, n);
+  for (i=0; i<n; i++) { v[i] = va_arg(a,GEN*); *v[i] = (GEN)copy_bin(*v[i]); }
+  set_avma(av);
+  for (i=0; i<n; i++) *v[i] = bin_copy((GENbin*)*v[i]);
+  return *v[0];
+}
+
 INLINE void
 gerepilecoeffs(pari_sp av, GEN x, int n)
 {
   int i;
   for (i=0; i<n; i++) gel(x,i) = (GEN)copy_bin(gel(x,i));
-  avma = av;
+  set_avma(av);
   for (i=0; i<n; i++) gel(x,i) = bin_copy((GENbin*)x[i]);
 }
 
@@ -1277,7 +1389,7 @@ INLINE void
 cgiv(GEN x)
 {
   pari_sp av = (pari_sp)(x+lg(x));
-  if (isonstack((GEN)av)) avma = av;
+  if (isonstack((GEN)av)) set_avma(av);
 }
 
 INLINE void
@@ -1319,10 +1431,24 @@ gtodouble(GEN x)
     pari_sp av = avma;
     x = gtofp(x, DEFAULTPREC);
     if (typ(x)!=t_REAL) pari_err_TYPE("gtodouble [t_REAL expected]", x);
-    avma = av;
+    set_avma(av);
   }
   return rtodbl(x);
 }
+
+INLINE int
+gisdouble(GEN x, double *g)
+{
+  if (typ(x)!=t_REAL) {
+    pari_sp av = avma;
+    x = gtofp(x, DEFAULTPREC);
+    if (typ(x)!=t_REAL) pari_err_TYPE("gisdouble [t_REAL expected]", x);
+    set_avma(av);
+  }
+  if (expo(x) >= 0x3ff) return 0;
+  *g = rtodbl(x); return 1;
+}
+
 INLINE long
 gtos(GEN x) {
   if (typ(x) != t_INT) pari_err_TYPE("gtos [integer expected]",x);
@@ -1373,7 +1499,7 @@ gtofp(GEN z, long prec)
       if (isintzero(a)) {
         GEN y = cgetg(3, t_COMPLEX);
         b = cxcompotor(b, prec);
-        gel(y,1) = real_0_bit(expo(b) - prec2nbits(prec));
+        gel(y,1) = real_0_bit(expo(b) - prec);
         gel(y,2) = b; return y;
       }
       return cxtofp(z, prec);
@@ -1455,7 +1581,7 @@ affgr(GEN x, GEN y)
     case t_INT:  affir(x,y); break;
     case t_REAL: affrr(x,y); break;
     case t_FRAC: rdiviiz(gel(x,1),gel(x,2), y); break;
-    case t_QUAD: av = avma; affgr(quadtofp(x,realprec(y)), y); avma = av; break;
+    case t_QUAD: av = avma; affgr(quadtofp(x,realprec(y)), y); set_avma(av); break;
     default: pari_err_TYPE2("=",x,y);
   }
 }
@@ -1470,7 +1596,7 @@ affc_fixlg(GEN x, GEN res)
   }
   else
   {
-    avma = (pari_sp)(res+3);
+    set_avma((pari_sp)(res+3));
     res = cgetr(realprec(gel(res,1)));
     affrr_fixlg(x, res);
   }
@@ -1488,7 +1614,7 @@ trunc_safe(GEN x) { long e; return gcvtoi(x,&e); }
 INLINE long
 ndec2nlong(long x) { return 1 + (long)((x)*(LOG2_10/BITS_IN_LONG)); }
 INLINE long
-ndec2prec(long x) { return 2 + ndec2nlong(x); }
+ndec2prec(long x) { return ndec2nlong(x) << TWOPOTBITS_IN_LONG; }
 INLINE long
 ndec2nbits(long x) { return ndec2nlong(x) << TWOPOTBITS_IN_LONG; }
 /* Fast implementation of ceil(x / (8*sizeof(long))); typecast to (ulong)
@@ -1501,16 +1627,16 @@ nbits2nlong(long x) {
 
 INLINE long
 nbits2extraprec(long x) {
-  return (long)(((ulong)x+BITS_IN_LONG-1) >> TWOPOTBITS_IN_LONG);
+  return (((ulong)x+BITS_IN_LONG-1)>>TWOPOTBITS_IN_LONG) << TWOPOTBITS_IN_LONG;
 }
 
-/* Fast implementation of 2 + nbits2nlong(x) */
 INLINE long
 nbits2prec(long x) {
-  return (long)(((ulong)x+3*BITS_IN_LONG-1) >> TWOPOTBITS_IN_LONG);
+  return (((ulong)x+BITS_IN_LONG-1)>>TWOPOTBITS_IN_LONG) << TWOPOTBITS_IN_LONG;
 }
+
 INLINE long
-nbits2lg(long x) {
+prec2lg(long x) {
   return (long)(((ulong)x+3*BITS_IN_LONG-1) >> TWOPOTBITS_IN_LONG);
 }
 /* ceil(x / sizeof(long)) */
@@ -1519,21 +1645,21 @@ nchar2nlong(long x) {
   return (long)(((ulong)x+sizeof(long)-1) >> (TWOPOTBITS_IN_LONG-3L));
 }
 INLINE long
-prec2nbits(long x) { return (x-2) * BITS_IN_LONG; }
+prec2nbits(long x) { return x; }
 INLINE double
 bit_accuracy_mul(long x, double y) { return (x-2) * (BITS_IN_LONG*y); }
 INLINE double
-prec2nbits_mul(long x, double y) { return (x-2) * (BITS_IN_LONG*y); }
+prec2nbits_mul(long x, double y) { return x * y; }
 INLINE long
-bit_prec(GEN x) { return prec2nbits(realprec(x)); }
+bit_prec(GEN x) { return realprec(x); }
 INLINE long
-bit_accuracy(long x) { return prec2nbits(x); }
+bit_accuracy(long x) { return (x-2) * BITS_IN_LONG; }
 INLINE long
-prec2ndec(long x) { return (long)prec2nbits_mul(x, LOG10_2); }
+prec2ndec(long x) { return (long)(x * LOG10_2); }
 INLINE long
-nbits2ndec(long x) { return (long)(x * LOG10_2); }
+nbits2ndec(long x) { return prec2ndec(x); }
 INLINE long
-precdbl(long x) {return (x - 1) << 1;}
+precdbl(long x) {return x << 1;}
 INLINE long
 divsBIL(long n) { return n >> TWOPOTBITS_IN_LONG; }
 INLINE long
@@ -1559,8 +1685,8 @@ Fp_add(GEN a, GEN b, GEN m)
   {
     GEN t = subii(p, m);
     s = signe(t);
-    if (!s) { avma = av; return gen_0; }
-    if (s < 0) { avma = (pari_sp)p; return p; }
+    if (!s) return gc_const(av, gen_0);
+    if (s < 0) return gc_const((pari_sp)p, p);
     if (cmpii(t, m) < 0) return gerepileuptoint(av, t); /* general case ! */
     p = remii(t, m);
   }
@@ -1569,6 +1695,12 @@ Fp_add(GEN a, GEN b, GEN m)
   return gerepileuptoint(av, p);
 }
 INLINE GEN
+Fp_double(GEN x, GEN N)
+{
+  GEN z = shifti(x, 1);
+  return cmpii(z, N) >= 0? subii(z, N): z;
+}
+INLINE GEN
 Fp_sub(GEN a, GEN b, GEN m)
 {
   pari_sp av=avma;
@@ -1583,7 +1715,7 @@ Fp_sub(GEN a, GEN b, GEN m)
   else
   {
     GEN t = addii(p, m);
-    if (!s) { avma = av; return gen_0; }
+    if (!s) return gc_const(av, gen_0);
     if (s > 0) return gerepileuptoint(av, t); /* general case ! */
     p = modii(t, m);
   }
@@ -1640,7 +1772,7 @@ Fp_mul(GEN a, GEN b, GEN m)
   GEN p; /*HACK: assume modii use <=lg(p)+(lg(m)<<1) space*/
   (void)new_chunk(lg(a)+lg(b)+(lg(m)<<1));
   p = mulii(a,b);
-  avma = av; return modii(p,m);
+  set_avma(av); return modii(p,m);
 }
 INLINE GEN
 Fp_sqr(GEN a, GEN m)
@@ -1649,7 +1781,7 @@ Fp_sqr(GEN a, GEN m)
   GEN p; /*HACK: assume modii use <=lg(p)+(lg(m)<<1) space*/
   (void)new_chunk((lg(a)+lg(m))<<1);
   p = sqri(a);
-  avma = av; return remii(p,m); /*Use remii: p >= 0 */
+  set_avma(av); return remii(p,m); /*Use remii: p >= 0 */
 }
 INLINE GEN
 Fp_mulu(GEN a, ulong b, GEN m)
@@ -1664,7 +1796,7 @@ Fp_mulu(GEN a, ulong b, GEN m)
     GEN p; /*HACK: assume modii use <=lg(p)+(lg(m)<<1) space*/
     (void)new_chunk(lg(a)+1+(l<<1));
     p = muliu(a,b);
-    avma = av; return modii(p,m);
+    set_avma(av); return modii(p,m);
   }
 }
 INLINE GEN
@@ -1686,7 +1818,7 @@ Fp_muls(GEN a, long b, GEN m)
     GEN p; /*HACK: assume modii use <=lg(p)+(lg(m)<<1) space*/
     (void)new_chunk(lg(a)+1+(l<<1));
     p = mulis(a,b);
-    avma = av; return modii(p,m);
+    set_avma(av); return modii(p,m);
   }
 }
 
@@ -1707,11 +1839,32 @@ Fp_invsafe(GEN a, GEN m)
 INLINE GEN
 Fp_div(GEN a, GEN b, GEN m)
 {
-  pari_sp av=avma;
-  GEN p; /*HACK: assume modii use <=lg(p)+(lg(m)<<1) space*/
+  pari_sp av = avma;
+  GEN p;
+  if (lgefint(b) == 3)
+  {
+    a = Fp_divu(a, b[2], m);
+    if (signe(b) < 0) a = Fp_neg(a, m);
+    return a;
+  }
+  /*HACK: assume modii use <=lg(p)+(lg(m)<<1) space*/
   (void)new_chunk(lg(a)+(lg(m)<<1));
   p = mulii(a, Fp_inv(b,m));
-  avma = av; return modii(p,m);
+  set_avma(av); return modii(p,m);
+}
+INLINE GEN
+Fp_divu(GEN x, ulong a, GEN p)
+{
+  pari_sp av = avma;
+  ulong b;
+  if (lgefint(p) == 3)
+  {
+    ulong pp = p[2], xp = umodiu(x, pp);
+    return xp? utoipos(Fl_div(xp, a % pp, pp)): gen_0;
+  }
+  x = Fp_red(x, p);
+  b = Fl_neg(Fl_div(umodiu(x,a), umodiu(p,a), a), a); /* x + pb = 0 (mod a) */
+  return gerepileuptoint(av, diviuexact(addmuliu(x, p, b), a));
 }
 
 INLINE GEN
@@ -1789,7 +1942,7 @@ submulii(GEN x, GEN y, GEN z)
   lz = lgefint(z);
   av = avma; (void)new_chunk(lx+ly+lz); /* HACK */
   t = mulii(z, y);
-  avma = av; return subii(x,t);
+  set_avma(av); return subii(x,t);
 }
 /* y*z - x */
 INLINE GEN
@@ -1804,7 +1957,7 @@ mulsubii(GEN y, GEN z, GEN x)
   lz = lgefint(z);
   av = avma; (void)new_chunk(lx+ly+lz); /* HACK */
   t = mulii(z, y);
-  avma = av; return subii(t,x);
+  set_avma(av); return subii(t,x);
 }
 
 /* x - u*y */
@@ -1817,7 +1970,7 @@ submuliu(GEN x, GEN y, ulong u)
   av = avma;
   (void)new_chunk(3+ly+lgefint(x)); /* HACK */
   y = mului(u,y);
-  avma = av; return subii(x, y);
+  set_avma(av); return subii(x, y);
 }
 /* x + u*y */
 INLINE GEN
@@ -1829,7 +1982,7 @@ addmuliu(GEN x, GEN y, ulong u)
   av = avma;
   (void)new_chunk(3+ly+lgefint(x)); /* HACK */
   y = mului(u,y);
-  avma = av; return addii(x, y);
+  set_avma(av); return addii(x, y);
 }
 /* x - u*y */
 INLINE GEN
@@ -1841,7 +1994,7 @@ submuliu_inplace(GEN x, GEN y, ulong u)
   av = avma;
   (void)new_chunk(3+ly+lgefint(x)); /* HACK */
   y = mului(u,y);
-  avma = av; return subii(x, y);
+  set_avma(av); return subii(x, y);
 }
 /* x + u*y */
 INLINE GEN
@@ -1853,7 +2006,7 @@ addmuliu_inplace(GEN x, GEN y, ulong u)
   av = avma;
   (void)new_chunk(3+ly+lgefint(x)); /* HACK */
   y = mului(u,y);
-  avma = av; return addii(x, y);
+  set_avma(av); return addii(x, y);
 }
 /* ux + vy */
 INLINE GEN
@@ -1868,7 +2021,7 @@ lincombii(GEN u, GEN v, GEN x, GEN y)
   av = avma; (void)new_chunk(lx+ly+lgefint(u)+lgefint(v)); /* HACK */
   p1 = mulii(u,x);
   p2 = mulii(v,y);
-  avma = av; return addii(p1,p2);
+  set_avma(av); return addii(p1,p2);
 }
 
 /*******************************************************************/
@@ -1888,6 +2041,8 @@ is_matvec_t(long t) { return (t >= t_VEC && t <= t_MAT); }
 INLINE int
 is_noncalc_t(long tx) { return (tx) >= t_LIST; }
 INLINE int
+is_qfb_t(long t) { return (t == t_QFB); }
+INLINE int
 is_rational_t(long t) { return (t == t_INT || t == t_FRAC); }
 INLINE int
 is_real_t(long t) { return (t == t_INT || t == t_REAL || t == t_FRAC); }
@@ -1898,6 +2053,9 @@ is_scalar_t(long t) { return (t < t_POL); }
 INLINE int
 is_vec_t(long t) { return (t == t_VEC || t == t_COL); }
 
+INLINE int
+qfb_is_qfi(GEN q) { return signe(gel(q,4)) < 0; }
+
 /*******************************************************************/
 /*                                                                 */
 /*                         TRANSCENDENTAL                          */
@@ -1989,6 +2147,13 @@ mul_content(GEN cx, GEN cy)
 INLINE GEN
 inv_content(GEN c) { return c? ginv(c): NULL; }
 INLINE GEN
+div_content(GEN cx, GEN cy)
+{
+  if (!cy) return cx;
+  if (!cx) return ginv(cy);
+  return gdiv(cx,cy);
+}
+INLINE GEN
 mul_denom(GEN dx, GEN dy)
 {
   if (!dx) return dy;
@@ -2050,6 +2215,9 @@ FpX_div(GEN x, GEN y, GEN p) { return FpX_divrem(x,y,p, NULL); }
 INLINE GEN
 Flx_div(GEN x, GEN y, ulong p) { return Flx_divrem(x,y,p, NULL); }
 INLINE GEN
+Flx_div_pre(GEN x, GEN y, ulong p, ulong pi)
+{ return Flx_divrem_pre(x,y,p,pi, NULL); }
+INLINE GEN
 F2x_div(GEN x, GEN y) { return F2x_divrem(x,y, NULL); }
 INLINE GEN
 FpV_FpC_mul(GEN x, GEN y, GEN p) { return FpV_dotproduct(x,y,p); }
@@ -2106,6 +2274,10 @@ INLINE GEN
 FpXQX_renormalize(GEN x, long lx) { return ZXX_renormalize(x,lx); }
 INLINE GEN
 F2x_renormalize(GEN x, long lx)   { return Flx_renormalize(x,lx); }
+
+INLINE GEN
+F2xX_shift(GEN a, long n, long vs) { return FlxX_shift(a, n, vs); }
+
 INLINE GEN
 F2v_to_F2x(GEN x, long sv) {
   GEN y = leafcopy(x);
@@ -2116,11 +2288,8 @@ INLINE long
 sturm(GEN x) { return sturmpart(x, NULL, NULL); }
 
 INLINE long
-gval(GEN x, long v) {
-  pari_sp av = avma;
-  long n = gvaluation(x, pol_x(v));
-  avma = av; return n;
-}
+gval(GEN x, long v)
+{ pari_sp av = avma; return gc_long(av, gvaluation(x, pol_x(v))); }
 
 INLINE void
 RgX_shift_inplace_init(long v)
@@ -2130,18 +2299,20 @@ INLINE GEN
 RgX_shift_inplace(GEN x, long v)
 {
   long i, lx;
-  GEN y, z;
+  GEN z;
   if (!v) return x;
   lx = lg(x);
   if (lx == 2) return x;
-  y = x + v;
   z = x + lx;
-  /* stackdummy from normalizepol: move it up */
-  if (lg(z) != v) x[lx + v] = z[0];
-  for (i = lx-1; i >= 2; i--) gel(y,i) = gel(x,i);
-  for (i = v+1;  i >= 2; i--) gel(x,i) = gen_0;
-  /* leave x[1] alone: it is correct */
-  x[0] = evaltyp(t_POL) | evallg(lx+v); return x;
+  /* stackdummy's from normalizepol */
+  while (lg(z) != v) z += lg(z);
+  z += v;
+  for (i = lx-1; i >= 2; i--) gel(--z,0) = gel(x,i);
+  for (i = 0;  i < v; i++) gel(--z,0) = gen_0;
+  z -= 2;
+  z[1] = x[1];
+  z[0] = evaltyp(t_POL) | evallg(lx+v);
+  stackdummy((pari_sp)z, (pari_sp)x); return z;
 }
 
 
@@ -2190,6 +2361,9 @@ RgM_shallowcopy(GEN x)
 INLINE GEN
 F2m_copy(GEN x) { return RgM_shallowcopy(x); }
 
+INLINE GEN
+F3m_copy(GEN x) { return RgM_shallowcopy(x); }
+
 INLINE GEN
 Flm_copy(GEN x) { return RgM_shallowcopy(x); }
 
@@ -2203,6 +2377,16 @@ ZV_dvd(GEN x, GEN y)
     if ( ! dvdii( gel(x,i), gel(y,i) ) ) return 0;
   return 1;
 }
+INLINE GEN
+ZM_ZV_mod(GEN x, GEN y)
+{ pari_APPLY_same(ZV_ZV_mod(gel(x,i), y)) }
+INLINE GEN
+ZV_ZV_mod(GEN x, GEN y)
+{ pari_APPLY_same(modii(gel(x,i), gel(y,i))) }
+INLINE GEN
+vecmodii(GEN x, GEN y) { return ZV_ZV_mod(x,y); }
+INLINE GEN
+vecmoduu(GEN x, GEN y) { pari_APPLY_ulong(((ulong*)x)[i] % ((ulong*)y)[i]) }
 
 /* Fq */
 INLINE GEN
@@ -2222,14 +2406,24 @@ gener_Fq_local(GEN T, GEN p, GEN L)
 { return T? gener_FpXQ_local(T,p, L)
           : pgener_Fp_local(p, L); }
 
+INLINE GEN
+random_Fq(GEN T, GEN p)
+{ return T ? random_FpX(get_FpX_degree(T), get_FpX_var(T), p): randomi(p); }
+
 /* FpXQX */
 INLINE GEN
 FpXQX_div(GEN x, GEN y, GEN T, GEN p) { return FpXQX_divrem(x, y, T, p, NULL); }
 INLINE GEN
 FlxqX_div(GEN x, GEN y, GEN T, ulong p) { return FlxqX_divrem(x, y, T, p, NULL); }
 INLINE GEN
+FlxqX_div_pre(GEN x, GEN y, GEN T, ulong p, ulong pi) { return FlxqX_divrem_pre(x, y, T, p, pi, NULL); }
+INLINE GEN
 F2xqX_div(GEN x, GEN y, GEN T) { return F2xqX_divrem(x, y, T, NULL); }
 
+INLINE GEN
+FpXY_Fq_evaly(GEN Q, GEN y, GEN T, GEN p, long vx)
+{ return T ? FpXY_FpXQ_evaly(Q, y, T, p, vx): FpXY_evaly(Q, y, p, vx); }
+
 /* FqX */
 INLINE GEN
 FqX_red(GEN z, GEN T, GEN p) { return T? FpXQX_red(z, T, p): FpX_red(z, p); }
@@ -2338,6 +2532,9 @@ FqXQ_pow(GEN x, GEN n, GEN S, GEN T, GEN p)
 
 /*FqXn*/
 INLINE GEN
+FqXn_expint(GEN x, long n, GEN T, GEN p)
+{ return T? FpXQXn_expint(x,n,T,p): FpXn_expint(x,n,p); }
+INLINE GEN
 FqXn_exp(GEN x, long n, GEN T, GEN p)
 { return T? FpXQXn_exp(x,n,T,p): FpXn_exp(x,n,p); }
 INLINE GEN
@@ -2424,6 +2621,14 @@ F2m_set(GEN x, long a, long b) { F2v_set(gel(x,b), a); }
 INLINE void
 F2m_flip(GEN x, long a, long b) { F2v_flip(gel(x,b), a); }
 
+/* F3m */
+
+INLINE ulong
+F3m_coeff(GEN x, long a, long b) { return F3v_coeff(gel(x,b), a); }
+
+INLINE void
+F3m_set(GEN x, long a, long b, ulong c) { F3v_set(gel(x,b), a, c); }
+
 /* ARITHMETIC */
 INLINE GEN
 matpascal(long n) { return matqpascal(n, NULL); }
@@ -2446,10 +2651,17 @@ gminsg(long s, GEN y) { return gmings(y,s); }
 INLINE GEN
 gmulgs(GEN y, long s) { return gmulsg(s,y); }
 INLINE GEN
+gmulgu(GEN y, ulong s) { return gmulug(s,y); }
+INLINE GEN
 gsubgs(GEN y, long s) { return gaddgs(y, -s); }
 INLINE GEN
 gdivsg(long s, GEN y) { return gdiv(stoi(s), y); }
 
+INLINE GEN
+gmax_shallow(GEN x, GEN y) { return gcmp(x,y)<0? y: x; }
+INLINE GEN
+gmin_shallow(GEN x, GEN y) { return gcmp(x,y)<0? x: y; }
+
 /* x t_COMPLEX */
 INLINE GEN
 cxnorm(GEN x) { return gadd(gsqr(gel(x,1)), gsqr(gel(x,2))); }
@@ -2481,7 +2693,7 @@ quad_disc(GEN x)
 INLINE GEN
 qfb_disc3(GEN x, GEN y, GEN z) { return subii(sqri(y), shifti(mulii(x,z),2)); }
 INLINE GEN
-qfb_disc(GEN x) { return qfb_disc3(gel(x,1), gel(x,2), gel(x,3)); }
+qfb_disc(GEN x) { return gel(x,4); }
 
 INLINE GEN
 sqrfrac(GEN x)
@@ -2529,33 +2741,33 @@ powIs(long n) {
 /*                                                                 */
 /*******************************************************************/
 INLINE void mpexpz(GEN x, GEN z)
-{ pari_sp av = avma; gaffect(mpexp(x), z); avma = av; }
+{ pari_sp av = avma; gaffect(mpexp(x), z); set_avma(av); }
 INLINE void mplogz(GEN x, GEN z)
-{ pari_sp av = avma; gaffect(mplog(x), z); avma = av; }
+{ pari_sp av = avma; gaffect(mplog(x), z); set_avma(av); }
 INLINE void mpcosz(GEN x, GEN z)
-{ pari_sp av = avma; gaffect(mpcos(x), z); avma = av; }
+{ pari_sp av = avma; gaffect(mpcos(x), z); set_avma(av); }
 INLINE void mpsinz(GEN x, GEN z)
-{ pari_sp av = avma; gaffect(mpsin(x), z); avma = av; }
+{ pari_sp av = avma; gaffect(mpsin(x), z); set_avma(av); }
 INLINE void gnegz(GEN x, GEN z)
-{ pari_sp av = avma; gaffect(gneg(x), z); avma = av; }
+{ pari_sp av = avma; gaffect(gneg(x), z); set_avma(av); }
 INLINE void gabsz(GEN x, long prec, GEN z)
-{ pari_sp av = avma; gaffect(gabs(x,prec), z); avma = av; }
+{ pari_sp av = avma; gaffect(gabs(x,prec), z); set_avma(av); }
 INLINE void gaddz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; gaffect(gadd(x,y), z); avma = av; }
+{ pari_sp av = avma; gaffect(gadd(x,y), z); set_avma(av); }
 INLINE void gsubz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; gaffect(gsub(x,y), z); avma = av; }
+{ pari_sp av = avma; gaffect(gsub(x,y), z); set_avma(av); }
 INLINE void gmulz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; gaffect(gmul(x,y), z); avma = av; }
+{ pari_sp av = avma; gaffect(gmul(x,y), z); set_avma(av); }
 INLINE void gdivz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; gaffect(gdiv(x,y), z); avma = av; }
+{ pari_sp av = avma; gaffect(gdiv(x,y), z); set_avma(av); }
 INLINE void gdiventz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; gaffect(gdivent(x,y), z); avma = av; }
+{ pari_sp av = avma; gaffect(gdivent(x,y), z); set_avma(av); }
 INLINE void gmodz(GEN x, GEN y, GEN z)
-{ pari_sp av = avma; gaffect(gmod(x,y), z); avma = av; }
+{ pari_sp av = avma; gaffect(gmod(x,y), z); set_avma(av); }
 INLINE void gmul2nz(GEN x, long s, GEN z)
-{ pari_sp av = avma; gaffect(gmul2n(x,s), z); avma = av; }
+{ pari_sp av = avma; gaffect(gmul2n(x,s), z); set_avma(av); }
 INLINE void gshiftz(GEN x, long s, GEN z)
-{ pari_sp av = avma; gaffect(gshift(x,s), z); avma = av; }
+{ pari_sp av = avma; gaffect(gshift(x,s), z); set_avma(av); }
 
 /*******************************************************************/
 /*                                                                 */
@@ -2606,7 +2818,7 @@ INLINE long pr_get_e(GEN pr) { return gel(pr,3)[2]; }
 INLINE long pr_get_f(GEN pr) { return gel(pr,4)[2]; }
 INLINE GEN pr_get_tau(GEN pr){ return gel(pr,5); }
 INLINE int
-pr_is_inert(GEN P) { return pr_get_f(P) == lg(pr_get_gen(P))-1; }
+pr_is_inert(GEN P) { return typ(pr_get_tau(P)) == t_INT; }
 INLINE GEN
 pr_norm(GEN pr) { return powiu(pr_get_p(pr), pr_get_f(pr)); }
 INLINE ulong
@@ -2676,6 +2888,8 @@ nf_get_sign(GEN nf, long *r1, long *r2)
   *r2 = itou(gel(x,2));
 }
 
+INLINE GEN
+cyc_get_expo(GEN c) { return lg(c) == 1? gen_1: gel(c,1); }
 INLINE GEN
 abgrp_get_no(GEN x) { return gel(x,1); }
 INLINE GEN
@@ -2697,17 +2911,23 @@ bnf_get_reg(GEN bnf) { return gmael(bnf,8,2); }
 INLINE GEN
 bnf_get_logfu(GEN bnf) { return gel(bnf,3); }
 INLINE GEN
+bnf_get_sunits(GEN bnf)
+{ GEN s = gmael(bnf,8,3); return typ(s) == t_INT? NULL: s; }
+INLINE GEN
 bnf_get_tuU(GEN bnf) { return gmael3(bnf,8,4,2); }
 INLINE long
 bnf_get_tuN(GEN bnf) { return gmael3(bnf,8,4,1)[2]; }
 INLINE GEN
-bnf_get_fu(GEN bnf) {
-  GEN fu = bnf_get_fu_nocheck(bnf);
-  if (typ(fu) == t_MAT) pari_err(e_MISC,"missing units in bnf");
-  return fu;
-}
-INLINE GEN
 bnf_get_fu_nocheck(GEN bnf) { return gmael(bnf,8,5); }
+INLINE GEN
+nfV_to_scalar_or_alg(GEN nf, GEN x)
+{ pari_APPLY_same(nf_to_scalar_or_alg(nf, gel(x,i))) }
+INLINE GEN
+bnf_get_fu(GEN bnf) {
+  GEN fu = bnf_build_units(bnf), nf = bnf_get_nf(bnf);
+  if (typ(fu) == t_MAT) pari_err(e_MISC,"missing units in bnf");
+  return nfV_to_scalar_or_alg(nf, vecslice(fu, 2, lg(fu)-1));
+}
 
 INLINE GEN
 bnr_get_bnf(GEN bnr) { return gel(bnr,1); }
@@ -2733,6 +2953,109 @@ bnr_get_gen(GEN bnr) {
   return gel(G,3);
 }
 
+/* localstar, used in gchar */
+INLINE GEN
+locs_get_cyc(GEN locs) { return gel(locs,1); }
+INLINE GEN
+locs_get_Lsprk(GEN locs) { return gel(locs,2); }
+INLINE GEN
+locs_get_Lgenfil(GEN locs) { return gel(locs,3); }
+INLINE GEN
+locs_get_mod(GEN locs) { return gel(locs,4); }
+/* pr dividing the modulus N of locs, 0 <= i < v_pr(N)
+ * return a t_MAT whose columns are the logs
+ * of generators of U_i(pr)/U_{i+1}(pr). */
+INLINE GEN
+locs_get_famod(GEN locs) { return gmael(locs,4,1); }
+INLINE GEN
+locs_get_m_infty(GEN locs) { return gmael(locs,4,2); }
+
+/* G a grossenchar group */
+INLINE GEN
+gchar_get_basis(GEN gc)  { return  gel(gc, 1); }
+INLINE GEN
+gchar_get_bnf(GEN gc)    { return  gel(gc, 2); }
+INLINE GEN
+gchar_get_nf(GEN gc)    { return  gel(gc, 3); }
+INLINE GEN
+gchar_get_zm(GEN gc)     { return  gel(gc, 4); }
+INLINE GEN
+gchar_get_mod(GEN gc)    { return  locs_get_mod(gchar_get_zm(gc)); }
+INLINE GEN
+gchar_get_modP(GEN gc)    { return gmael(gchar_get_mod(gc),1,1); }
+INLINE GEN
+gchar_get_S(GEN gc)      { return  gel(gc, 5); }
+INLINE GEN
+gchar_get_DLdata(GEN gc)   { return  gel(gc, 6); }
+INLINE GEN
+gchar_get_sfu(GEN gc) { return  gel(gc, 7); }
+INLINE GEN
+gchar_get_cyc(GEN gc)    { return  gel(gc, 9); }
+INLINE GEN
+gchar_get_hnf(GEN gc)    { return  gmael(gc, 10, 1); }
+INLINE GEN
+gchar_get_U(GEN gc)      { return  gmael(gc, 10, 2); }
+INLINE GEN
+gchar_get_Ui(GEN gc)     { return  gmael(gc, 10, 3); }
+INLINE GEN
+gchar_get_m0(GEN gc)     { return gel(gc, 11); }
+INLINE GEN
+gchar_get_u0(GEN gc)     { return gel(gc, 12); }
+INLINE long
+gchar_get_r1(GEN gc)     { return nf_get_r1(gchar_get_nf(gc)); }
+INLINE long
+gchar_get_r2(GEN gc)     { return nf_get_r2(gchar_get_nf(gc)); }
+INLINE GEN
+gchar_get_loccyc(GEN gc) { return locs_get_cyc(gchar_get_zm(gc)); }
+INLINE long
+gchar_get_nc(GEN gc)     { return lg(gchar_get_loccyc(gc))-1; }
+INLINE long
+gchar_get_ns(GEN gc)     { return lg(gchar_get_S(gc))-1; }
+INLINE long
+gchar_get_nm(GEN gc)     { return lg(gchar_get_basis(gc))-1; }
+INLINE long
+gchar_get_evalprec(GEN gc)   { return  gmael(gc, 8, 1)[1]; }
+INLINE long
+gchar_get_prec(GEN gc)   { return  gmael(gc, 8, 1)[2]; }
+INLINE long
+gchar_get_nfprec(GEN gc) { return  gmael(gc, 8, 1)[3]; }
+INLINE void
+gchar_set_evalprec(GEN gc, long prec) { gmael(gc, 8, 1)[1] = prec; }
+INLINE void
+gchar_set_prec(GEN gc, long prec) { gmael(gc, 8, 1)[2] = prec; }
+INLINE void
+gchar_copy_precs(GEN gc, GEN gc2)
+{
+  gel(gc2, 8) = shallowcopy(gel(gc,8));
+  gmael(gc2, 8, 1) = shallowcopy(gmael(gc, 8, 1));
+}
+INLINE void
+gchar_set_nfprec(GEN gc, long prec) { gmael(gc, 8, 1)[3] = prec; }
+INLINE long
+gchar_get_ntors(GEN gc)   { return  gmael(gc, 8, 2)[1]; }
+INLINE long
+gchar_get_nfree(GEN gc)   { return  gmael(gc, 8, 2)[2]; }
+INLINE long
+gchar_get_nalg(GEN gc)   { return  gmael(gc, 8, 2)[3]; }
+INLINE void
+gchar_set_basis(GEN gc, GEN m_inv)  { gel(gc, 1) = m_inv; }
+INLINE void
+gchar_set_nf(GEN gc, GEN nf)      { gel(gc, 3) = nf; }
+INLINE void
+gchar_set_ntors(GEN gc, long n)    { gmael(gc, 8, 2)[1] = n; }
+INLINE void
+gchar_set_nfree(GEN gc, long n)    { gmael(gc, 8, 2)[2] = n; }
+INLINE void
+gchar_set_nalg(GEN gc, long n)    { gmael(gc, 8, 2)[3] = n; }
+INLINE void
+gchar_set_cyc(GEN gc, GEN cyc)      { gel(gc, 9) = cyc; }
+INLINE void
+gchar_set_HUUi(GEN gc, GEN hnf, GEN U, GEN Ui) { gel(gc, 10) = mkvec3(hnf, U, Ui); }
+INLINE void
+gchar_set_m0(GEN gc, GEN m0)     { gel(gc, 11) = m0; }
+INLINE void
+gchar_set_u0(GEN gc, GEN u0)     { gel(gc, 12) = u0; }
+
 INLINE GEN
 bid_get_mod(GEN bid) { return gel(bid,1); }
 INLINE GEN
@@ -2746,6 +3069,8 @@ bid_get_fact(GEN bid) { return gmael(bid,3,1); }
 INLINE GEN
 bid_get_fact2(GEN bid) { return gmael(bid,3,2); }
 INLINE GEN
+bid_get_MOD(GEN bid) { GEN y = gel(bid,4); return lg(y) == 4? gel(y,3): NULL; }
+INLINE GEN
 bid_get_sprk(GEN bid) { return gmael(bid,4,1); }
 INLINE GEN
 bid_get_sarch(GEN bid) { return gmael(bid,4,2); }
@@ -2832,6 +3157,8 @@ INLINE GEN
 rnf_get_disc(GEN rnf) { return gel(rnf,3); }
 INLINE GEN
 rnf_get_index(GEN rnf) { return gel(rnf,4); }
+INLINE GEN
+rnf_get_ramified_primes(GEN rnf) { return gel(rnf,5); }
 INLINE long
 rnf_get_varn(GEN rnf) { return varn(gel(rnf,1)); }
 INLINE GEN
@@ -2845,38 +3172,8 @@ rnf_get_map(GEN rnf) { return gel(rnf,11); }
 INLINE GEN
 rnf_get_invzk(GEN rnf) { return gel(rnf,8); }
 
-/* I integral ZM (not HNF), G ZM, rounded Cholesky form of a weighted
- * T2 matrix. Reduce I wrt G */
 INLINE GEN
-idealpseudored(GEN I, GEN G)
-{ return ZM_mul(I, ZM_lll(ZM_mul(G, I), 0.99, LLL_IM)); }
-
-/* I integral (not necessarily HNF), G ZM, rounded Cholesky form of a weighted
- * T2 matrix. Return m in I with T2(m) small */
-INLINE GEN
-idealpseudomin(GEN I, GEN G)
-{
-  GEN u = ZM_lll(ZM_mul(G, I), 0.99, LLL_IM);
-  return ZM_ZC_mul(I, gel(u,1));
-}
-/* I, G as in idealpseudomin. Return an irrational m in I with T2(m) small */
-INLINE GEN
-idealpseudomin_nonscalar(GEN I, GEN G)
-{
-  GEN u = ZM_lll(ZM_mul(G, I), 0.99, LLL_IM);
-  GEN m = ZM_ZC_mul(I, gel(u,1));
-  if (ZV_isscalar(m) && lg(u) > 2) m = ZM_ZC_mul(I, gel(u,2));
-  return m;
-}
-
-INLINE GEN
-idealred_elt(GEN nf, GEN I) {
-  pari_sp av = avma;
-  GEN u = idealpseudomin(I, nf_get_roundG(nf));
-  return gerepileupto(av, u);
-}
-INLINE GEN
-idealred(GEN nf, GEN I) { return idealred0(nf, I, NULL); }
+idealred(GEN nf, GEN id) { return idealred0(nf, id, NULL); }
 
 INLINE GEN
 idealchineseinit(GEN nf, GEN x)
@@ -2958,3 +3255,4 @@ pari_err_VAR(const char *f, GEN x, GEN y) { pari_err(e_VAR, f,x,y); }
 INLINE void
 pari_err_PRIORITY(const char *f, GEN x, const char *op, long v)
 { pari_err(e_PRIORITY, f,x,op,v); }
+
